class FatalError(Exception):
    """Exception wrapper that contains an exit status in addition to a message"""
    def __init__(self, message, status=1):
        self.message = message
        self.status = status



# Copyright 2015 Johannes Grassler <johannes@btw23.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re

class AnchorFilter(object):
    """Strips out HTML anchor tags"""

    def run(self, lines):
        """Filter method"""
        ret = []
        for line in lines:
            ret.append(re.sub(r'<a.*?</a>', '', line))

        return ret



#!/usr/bin/python
#
# Copyright 2015 Johannes Grassler <johannes@btw23.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# mkdocs2pandoc - converts mkdocs documentation into a single pandoc markdown document

from __future__ import print_function

import argparse
import codecs
import sys

import mkdocs.config
from mkdocs_pandoc.exceptions import FatalError

import mkdocs_pandoc

def main():
    opts = argparse.ArgumentParser(
            description="mdtableconv.py " +
            "- converts pipe delimited tables to Pandoc's grid tables")

    opts.add_argument('-e', '--encoding', default='utf-8',
            help="Set encoding for input files (default: utf-8)")

    opts.add_argument('-f', '--config-file', default='mkdocs.yml',
            help="mkdocs configuration file to use")

    opts.add_argument('-i', '--image-ext', default=None,
            help="Extension to substitute image extensions by (default: no replacement)")

    opts.add_argument('-w', '--width', default=100,
            help="Width of generated grid tables in characters (default: 100)")

    opts.add_argument('-x', '--exclude', default=None, action='append',
            help="Include files to skip (default: none)")

    opts.add_argument('-o', '--outfile', default=None,
            help="File to write finished pandoc document to (default: STDOUT)")

    args = opts.parse_args()

    # Python 2 and Python 3 have mutually incompatible approaches to writing
    # encoded data to sys.stdout, so we'll have to pick the appropriate one.

    if sys.version_info.major == 2:
      out = codecs.getwriter(args.encoding)(sys.stdout)
    elif sys.version_info.major >= 3:
      out = open(sys.stdout.fileno(), mode='w', encoding=args.encoding, buffering=1)

    try:
      pconv = mkdocs_pandoc.PandocConverter(
                config_file=args.config_file,
                exclude=args.exclude,
                image_ext=args.image_ext,
                width=args.width,
                encoding=args.encoding,
                )
    except FatalError as e:
        print(e.message, file=sys.stderr)
        return(e.status)
    if args.outfile:
        try:
          out = codecs.open(args.outfile, 'w', encoding=args.encoding)
        except IOError as e:
          print("Couldn't open %s for writing: %s" % (args.outfile, e.strerror), file=sys.stderr)

    for line in pconv.convert():
        out.write(line + '\n')
    out.close()



#!/usr/bin/python
#
# Copyright 2015 Johannes Grassler <johannes@btw23.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# mdtableconv.py - converts pipe tables to Pandoc's grid tables

import markdown.extensions.tables as tbl
import markdown.blockparser
import re
import string
import textwrap

class TableFilter(tbl.TableProcessor):
    def __init__(self, width=100, encoding='utf-8'):
        self.width = width
        self.width_default = 20   # Default column width for rogue rows with more cells than the first row.


    def blocks(self, lines):
        """Groups lines into markdown blocks"""
        state = markdown.blockparser.State()
        blocks = []

        # We use three states: start, ``` and '\n'
        state.set('start')

        # index of current block
        currblock = 0

        for line in lines:
            line += '\n'
            if state.isstate('start'):
                if line[:3] == '```':
                    state.set('```')
                else:
                    state.set('\n')
                blocks.append('')
                currblock = len(blocks) - 1
            else:
                marker = line[:3]  # Will capture either '\n' or '```'
                if state.isstate(marker):
                    state.reset()
            blocks[currblock] += line

        return blocks


    def convert_table(self, block):
        """"Converts a table to grid table format"""
        lines_orig = block.split('\n')
        lines_orig.pop() # Remove extra newline at end of block
        widest_cell = [] # Will hold the width of the widest cell for each column
        widest_word = [] # Will hold the width of the widest word for each column
        widths = []      # Will hold the computed widths of grid table columns

        rows = []   # Will hold table cells during processing
        lines = []  # Will hold the finished table

        has_border = False  # Will be set to True if this is a bordered table

        width_unit = 0.0 # This number is used to divide up self.width according
                         # to the following formula:
                         #
                         # self.width = width_unit * maxwidth
                         #
                         # Where maxwidth is the sum over all elements of
                         # widest_cell.

        # Only process tables, leave everything else untouched

        if not self.test(None, block):
            return lines_orig

        if lines_orig[0].startswith('|'):
            has_border = True

        # Initialize width arrays

        for i in range(0, len(self._split_row(lines_orig[0], has_border))):
            widest_cell.append(0)
            widest_word.append(0)
            widths.append(0)

        # Parse lines into array of cells and record width of widest cell/word

        for line in lines_orig:
            row = self._split_row(line, has_border)
            for i in range(0, len(row)):
                # Record cell width
                if len(row[i]) > widest_cell[i]:
                    widest_cell[i] = len(row[i])
                # Record longest word
                words = row[i].split()
                for word in words:
                    # Keep URLs from throwing the word length count off too badly.
                    match = re.match(r'\[(.*?)\]\(.*?\)', word)
                    if match:
                       word = match.group(1)

                    if len(word) > widest_word[i]:
                        widest_word[i] = len(word)
            rows.append(row)

        # Remove table header divider line from rows
        rows.pop(1)

        # Compute first approximation of column widths based on maximum cell width

        for width in widest_cell:
            width_unit += float(width)

        width_unit = self.width / width_unit

        for i in range(0, len(widest_cell)):
            widths[i] = int(widest_cell[i] * width_unit)

        # Add rounding errors to narrowest column
        if sum(widths) < self.width:
            widths[widths.index(min(widths))] += self.width - sum(widths)

        # Attempt to correct first approximation of column widths based on
        # words that fail to fit their cell's width (if this fails textwrap
        # will break up long words but since it does not add hyphens this
        # should be avoided)

        for i in range(0, len(widths)):
            if widths[i] < widest_word[i]:
                offset = widest_word[i] - widths[i]
                for j in range(0, len(widths)):
                    if widths[j] - widest_word[j] >= offset:
                        widths[j] -= offset
                        widths[i] += offset
                        offset = 0

        lines.append(self.ruler_line(widths, linetype='-'))

        # Only add header row if it contains more than just whitespace
        if ''.join(rows[0]).strip() != '':
                lines.extend(self.wrap_row(widths, rows[0]))
                lines.append(self.ruler_line(widths, linetype='='))

        for row in rows[1:]:
            # Skip empty rows
            if ''.join(row).strip() == '':
                continue
            lines.extend(self.wrap_row(widths, row))
            lines.append(self.ruler_line(widths, linetype='-'))

        # Append empty line after table
        lines.append('')

        return lines


    def run(self, lines):
        """Filter method: Passes all blocks through convert_table() and returns a list of lines."""
        ret = []

        for block in self.blocks(lines):
            ret.extend(self.convert_table(block))

        return ret


    def ruler_line(self, widths, linetype='-'):
        """Generates a ruler line for separating rows from each other"""
        cells = []
        for w in widths:
            cells.append(linetype * (w+2))
        return '+' + '+'.join(cells) + '+'


    def wrap_row(self, widths, row, width_default=None):
        """Wraps a single line table row into a fixed width, multi-line table."""
        lines = []
        longest = 0 # longest wrapped column in row

        if not width_default:
            width_default = self.width_default

        # Wrap column contents
        for i in range(0, len(row)):
            w=width_default # column width

            # Only set column width dynamicaly for non-rogue rows
            if i < len(widths):
              w = widths[i]

            tw = textwrap.TextWrapper(width=w, break_on_hyphens=False)
            # Wrap and left-justify
            row[i] = tw.wrap(textwrap.dedent(row[i]))
            # Pad with spaces up to to fixed column width
            for l in range(0, len(row[i])):
                    row[i][l] += (w - len(row[i][l])) * ' '
            if len(row[i]) > longest:
                longest = len(row[i])

        # Pad all columns to have the same number of lines
        for i in range(0, len(row)):
            w=width_default # column width

            # Only set column width dynamicaly for non-rogue rows
            if i < len(widths):
                w = widths[i]

            if len(row[i]) < longest:
                for j in range(len(row[i]), longest):
                    row[i].append(w * ' ')

        for l in range(0,longest):
            line = []
            for c in range(len(row)):
                line.append(row[c][l])
            line = '| ' + ' | '.join(line) + ' |'
            lines.append(line)

        return lines



import mkdocs_pandoc.filters.anchors
import mkdocs_pandoc.filters.chapterhead
import mkdocs_pandoc.filters.headlevels
import mkdocs_pandoc.filters.images
import mkdocs_pandoc.filters.exclude
import mkdocs_pandoc.filters.include
import mkdocs_pandoc.filters.tables
import mkdocs_pandoc.filters.toc
import mkdocs_pandoc.filters.xref

from mkdocs_pandoc.exceptions import FatalError

import codecs
import os
import yaml


class PandocConverter:
    """Top level converter class. Instatiate separately for each mkdocs.yml."""

    def __init__(self, **kwargs):
        self.config_file = kwargs.get('config_file', 'mkdocs.yml')
        self.encoding = kwargs.get('encoding', 'utf-8')
        self.exclude = kwargs.get('exclude', None)
        self.filter_tables = kwargs.get('filter_tables', True)
        self.filter_xrefs = kwargs.get('filter_xrefs', True)
        self.image_ext = kwargs.get('image_ext', None)
        self.strip_anchors = kwargs.get('strip_anchors', True)
        self.width = kwargs.get('width', 100)

        try:
            cfg = codecs.open(self.config_file, 'r', self.encoding)
        except IOError as e:
            raise FatalError("Couldn't open %s for reading: %s" % (self.config_file,
                e.strerror), 1)

        self.config = yaml.load(cfg)

        if not 'docs_dir' in self.config:
            self.config['docs_dir'] = 'docs'

        if not 'site_dir' in  self.config:
            self.config['site_dir'] = 'site'

        # Set filters depending on markdown extensions from config
        # Defaults first...
        self.filter_include = False
        self.filter_toc = False

        # ...then override defaults based on config, if any:

        if 'markdown_extensions' in self.config:
          for ext in self.config['markdown_extensions']:
              extname = ''
              # extension entries may be dicts (for passing extension parameters)
              if type(ext) is dict:
                  extname = list(ext.keys())[0]
              if type(ext) is str:
                  extname = ext

              if extname == 'markdown_include.include':
                  self.filter_include = True
              if extname == 'toc':
                  self.filter_toc = True

        cfg.close()

    def flatten_pages(self, pages, level=1):
        """Recursively flattens pages data structure into a one-dimensional data structure"""
        flattened = []

        for page in pages:
            if type(list(page.values())[0]) is str:
                flattened.append(
                        {
                            'file': list(page.values())[0],
                            'title': list(page.keys())[0],
                            'level': level,
                         })
            if type(list(page.values())[0]) is list:
                flattened.extend(
                        self.flatten_pages(
                            list(page.values())[0],
                            level + 1)
                        )


        return flattened

    def convert(self):
        """User-facing conversion method. Returns pandoc document as a list of
        lines."""
        lines = []

        pages = self.flatten_pages(self.config['pages'])

        f_exclude = mkdocs_pandoc.filters.exclude.ExcludeFilter(
                exclude=self.exclude)

        f_include = mkdocs_pandoc.filters.include.IncludeFilter(
                base_path=self.config['docs_dir'],
                encoding=self.encoding)

        # First, do the processing that must be done on a per-file basis:
        # Adjust header levels, insert chapter headings and adjust image paths.

        f_headlevel = mkdocs_pandoc.filters.headlevels.HeadlevelFilter(pages)

        for page in pages:
            fname = os.path.join(self.config['docs_dir'], page['file'])
            try:
                p = codecs.open(fname, 'r', self.encoding)
            except IOError as e:
                raise FatalError("Couldn't open %s for reading: %s" % (fname,
                    e.strerror), 1)
            f_chapterhead = mkdocs_pandoc.filters.chapterhead.ChapterheadFilter(
                    headlevel=page['level'],
                    title=page['title']
                    )

            f_image = mkdocs_pandoc.filters.images.ImageFilter(
                    filename=page['file'],
                    image_path=self.config['site_dir'],
                    image_ext=self.image_ext)

            lines_tmp = []

            for line in p.readlines():
                lines_tmp.append(line.rstrip())

            if self.exclude:
                lines_tmp = f_exclude.run(lines_tmp)

            if self.filter_include:
                lines_tmp = f_include.run(lines_tmp)

            lines_tmp = f_headlevel.run(lines_tmp)
            lines_tmp = f_chapterhead.run(lines_tmp)
            lines_tmp = f_image.run(lines_tmp)
            lines.extend(lines_tmp)
            # Add an empty line between pages to prevent text from a previous
            # file from butting up against headers in a subsequent file.
            lines.append('')

        # Strip anchor tags
        if self.strip_anchors:
            lines = mkdocs_pandoc.filters.anchors.AnchorFilter().run(lines)

        # Fix cross references
        if self.filter_xrefs:
            lines = mkdocs_pandoc.filters.xref.XrefFilter().run(lines)

        if self.filter_toc:
            lines = mkdocs_pandoc.filters.toc.TocFilter().run(lines)

        if self.filter_tables:
            lines = mkdocs_pandoc.filters.tables.TableFilter().run(lines)

        return(lines)
